#!/usr/bin/perl -sw
##
## Razor2::Client::Core - Vipul's Razor Client API
##
## Copyright (c) 2001, Vipul Ved Prakash.  All rights reserved.
## This code is free software; you can redistribute it and/or modify
## it under the same terms as Perl itself.
##
## $Id: Core.pm,v 1.40 2002/06/28 03:30:19 chad Exp $

package Razor2::Client::Core;


    use strict;
    use IO::Socket;
    use IO::Select;
    use Errno qw(:POSIX);

    use Razor2::Client::Version; 
    use vars qw( $VERSION $PROTOCOL );
    use base qw(Razor2::String);
    use base qw(Razor2::Logger);
    use base qw(Razor2::Client::Engine);
    use base qw(Razor2::Errorhandler);
    use Razor2::String qw(hextobase64 makesis parsesis hmac_sha1 xor_key 
                          prep_mail debugobj to_batched_query 
                          from_batched_query hexbits2hash);



($VERSION) = do { my @r = (q$Revision: 1.40 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };
$PROTOCOL = $Razor2::Client::Version::PROTOCOL; 


sub new { 
    my ($class, $conf, %params) = @_;
    my $self = {};
    bless $self, $class;
    $self->debug ("Razor Agents $VERSION, protocol version $PROTOCOL.");
    return $self;
} 



#
# We store server-specific config info for each server we know about.
# All info about razor servers is stored in $self->{s}.
#
# Basically we get the server name/ip from {list}, 
# load that server's specific info from {allconfs} into {conf},
# and do stuff.  If server is no good, we get nextserver from {list}
#
# $self->{s}->{list}      ptr to {nsl} if report,revoke; or {csl} if check
#                         or the cmd-line server (-rs server)
# $self->{s}->{csl}       array ref containing catalogue servers
# $self->{s}->{nsl}       array ref containing nomination servers
# $self->{s}->{ds}        array ref containing discovery servers 
#
# $self->{s}->{modified}  array ref containing servers whose .conf needs updating
#
# $self->{s}->{ip}        string containing ip (or dns name) of current server from {list})
# $self->{s}->{port}      string containing port, taken from server:port from {list}
# $self->{s}->{engines}   engines supported, derived from {conf}->{se}
# $self->{s}->{conf}      hash ref containing current server's config params
#                         read from $razorhome/server.$ip.conf
#
# $self->{s}->{allconfs}  hash ref of all servers' configs.  key={ip}, val={conf}
#                         as read from server.*.conf file
#
# $self->{s}->{listfile}  string containing path/file of server.lst, either 
#                         nomination or catalogue depending $self->{breed}
#
# NOTE: if we are razor-check, server is Catalogue Server
#       otherwise server is Nomination server.
#
# everytime we update our server list, $self->{s}->{list};
# we want to write that to disk - $self->{s}->{listfile}  
#
sub nextserver {
    my ($self) = @_;

    # see if we need to discover (.lst files might be too old)
    $self->discover() or return $self->errprefix ("nextserver");

    # first time we don't remove from list
    shift @{$self->{s}->{list}} if $self->{s}->{ip}; 

    my $next = ${$self->{s}->{list}}[0];

    # do we ever want to put current back on the end of list?
    # push @{$self->{s}->{list}}, $self->{s}->{ip}; 

    if ($next) {
        ($self->{s}->{port}) = $next =~ /:(.*)$/;
        $next =~ s/:.*$//;         # optional
        $self->{s}->{ip}   = $next;  # ip can be IP or DNS name
        $self->{s}->{port} ||= $self->{conf}->{port};
        $self->{s}->{conf} = $self->{s}->{allconfs}->{$next};

        my $svrport = "$self->{s}->{ip}:$self->{s}->{port}";

        # get rid of server specific stuff
        delete $self->{s}->{greeting};

        unless (ref($self->{s}->{conf})) {
            # never used this server before, no cached info. go get it!
            $self->{s}->{conf} = {};
            $self->connect;   # calls parse_greeting which calls compute_server_conf
        } else {
            $self->compute_server_conf(1);  # computes supported engines, logs info
        }
        $self->writeservers();

        $self->log(8, "Using next closest server $svrport, cached info srl ". $self->{s}->{conf}->{srl});
        #$self->logobj(11, "Using next closest server $svrport, cached info", $self->{s}->{conf});
        return 1;

    } else {
        return $self->error ("Razor server $self->{opt}->{server} not available at this time")
            if $self->{opt}->{server};
        $self->{force_discovery} = 1;
        if ($self->{done_discovery} && !($self->discover)) {
            return $self->errprefix("No Razor servers available at this time");
        }
        return $self->nextserver;
    }
}

sub load_at_runtime {
    my ($self,$class,$sub,$args) = @_;

    eval "use $class";
    if ($@) {
        $self->log(2,"$class not found, please to fix.");
        return $self->error("\n\n$@");
    }
    my $evalstr;
    if ($sub && $sub ne "new") {
        $evalstr = $class ."::$sub($args);";
    } else {
        $evalstr = $class . "->new($args)"; 
    }
    if (my $dude = eval $evalstr) {
        $self->log(12,"Found and evaled $evalstr ==> $dude");
        return $dude;
    } else {
        $self->log(5,"Found but problem (bad args?) with $evalstr");
        return $self->error("Problem with $evalstr");
    }
}

# 
# uses DNS to find Discovery servers
# puts discovery servers in $self->{s}->{ds}
#
sub bootstrap_discovery {
    my ($self) = @_;
    
    if ($self->{conf}->{server}) {
        $self->log(8,"no bootstap_discovery when cmd-line server specified");
        return 1;
    }
    unless ($self->{force_bootstrap_discovery}) {
        if (ref($self->{conf}->{ds}) && scalar(@{$self->{conf}->{ds}})) {
            $self->log(8,"already have ". scalar(@{$self->{conf}->{ds}})
                ." discovery servers");
            return 1;
        } elsif ($self->{done_bootstrap}) {
            # if we've done it before {ds} should be set
            $self->log(8,"already have done bootstrap_discovery");
            return 1;
        }
    }

    my $zone = $self->{conf}->{razorzone};
    $self->log(5,"Finding Discovery Servers via DNS in the $zone zone");

    my $resolver = $self->load_at_runtime("Net::DNS::Resolver")
        or return $self->errprefix("bootstrap_discovery");

    my @list;

    BUILDLIST: 
    for ( "a".."z", "1..9999" ) { 
        my $query = $resolver->query ($_.'.'.$zone);
        my $rr; next BUILDLIST unless $query;
        foreach $rr ($query->answer) { 
            my $pushed = 0;
            if ($rr->type eq "A") { 
                push @list, $rr->address; 
                $pushed = 1;
            } elsif ($rr->type eq "CNAME") { 
                if ($rr->cname eq 'list.terminator') { 
                    pop @list if $pushed;
                    last BUILDLIST;
                } elsif ($rr->cname eq "skip") { 
                    pop @list if $pushed;
                    next BUILDLIST;
                }
            }
        }
    }
    $self->{done_bootstrap} = 1;

    $self->log(6,"Found ". scalar(@list) ." Discovery Servers via DNS in the $zone zone");

    return $self->error("No Razor Discovery servers available at this time")
        unless scalar(@list);

    $self->{s}->{ds} = \@list;
    $self->{s}->{ds_modified} = 1;
    return 1;
}



# 
# uses Discovery Servers to find closest Nomination/Catalogue Servers.
# called every day or so of if .lst file is empty
#
# puts servers in $self->{s}->{list}
#
sub discover {
    my ($self) = @_; 

    #
    # do we need to discover?
    #

    # no discover if cmd-line server
    return 1 if $self->{opt}->{server};

    #
    # don't discover if conf says turn_off_discovery (unless force_discovery)
    #
    return 1 if $self->{conf}->{turn_off_discovery} && (!($self->{force_discovery}));
    return $self->error ("No Razor servers available at this time")
        if $self->{done_discovery};

    # so if user has their own servers, and they are temporarily down, force_discovery.
    # good: shit will work
    #  bad: it will erase their custom server*.lst file
    #
    if (-s $self->{s}->{listfile}) {
        my $randomize = int(rand($self->{conf}->{rediscovery_wait}/7));
        my $timeleft = ((stat ($self->{s}->{listfile}))[9] + $self->{conf}->{rediscovery_wait} - $randomize) - time;
        if ($timeleft > 0) {
            $self->debug ("$timeleft seconds before closest server discovery");
            return 1 unless $self->{force_discovery};
            $self->debug ("forcing discovery");
        } else { 
            $self->debug ("server discovery overdue by ". (0-$timeleft) ." seconds"); 
        }
    } else { 
        if (-e) {
            $self->debug ("empty listfile: $self->{s}->{listfile}");
        } else { 
            $self->debug ("no listfile: $self->{s}->{listfile}");
        }
    } 

    #
    # we need to discover.
    #


    return $self->errprefix("discover0") unless $self->bootstrap_discovery();


    #
    # Go ahead and do discovery for both csl and nsl.
    #
    foreach my $querytype (qw(csl nsl)) {
        my %srvs;

    my $query = "a=g&pm=$querytype\r\n";
    foreach (@{$self->{s}->{ds}}) {

        $self->log (8,"Checking with Razor Discovery Server $_");

        unless ($self->connect( server => $_, discovery_server => 1 ) ) {
            $self->log (5,"Razor Discovery Server $_ not answering");
            next;
        }
        my $resp = $self->_send([$query]);
        unless ($resp) {
            return $self->errprefix("discover1");
            next;
        }

        # from_batched_query wants "-" in beginning, but not ".\r\n" at end
        $resp->[0] =~ s/\.\r\n$//sg;   
        my $h = from_batched_query($resp->[0], {});
        # $self->logobj(11,"response from $_", $h);

        foreach my $href (@$h) {
            next unless $href->{$querytype};
            $self->log (6,"Discovery Server $_ replying with $querytype=$href->{$querytype}");
            $srvs{$href->{$querytype}} = 1;
        }
        unless (keys %srvs) {
            $self->log (5,"Razor Discovery Server $_ had no valid servers");
            next;
        }
    }
    my @list = keys %srvs;

    return $self->error("Could not get valid info from Discovery Servers")
        unless @list;

        # $self->logobj(11,"checking ping times for servers", \@list);

    # sort @list by ping time
    #
    my $pinger = $self->load_at_runtime("Net::Ping", "new", "'tcp', 4")
        or return $self->errprefix("discover2");

    my %times; my $timings;
   
    for (@list) {
        my $t1 = $self->load_at_runtime("Time::HiRes", "gettimeofday");
        $pinger->ping($_);
        my $t2 = $self->load_at_runtime("Time::HiRes", "gettimeofday");
        $times{$_} = $t2 - $t1;
    }
    @list = sort { $times{$a} <=> $times{$b} } keys %times;
    for (@list) { $timings .= sprintf ("  %4.1fms %s,", $times{$_}*1000, $_);  }
    $self->log (5,"Sorted (closest first) list of $querytype servers & RTTs:$timings");
    $self->{s}->{$querytype} = \@list;
    $self->{s}->{$querytype ."_modified"} = 1;

    }

    unless ($self->{opt}->{server}) {
        if ($self->{breed} =~ /^check/) {
            $self->{s}->{list} = $self->{s}->{csl};
            $self->{s}->{listfile} = $self->{conf}->{listfile_catalogue}; # for discovery()
        } else {
            $self->{s}->{list} = $self->{s}->{nsl};
            $self->{s}->{listfile} = $self->{conf}->{listfile_nomination}; # for discovery()
        }
    }

    $self->{done_discovery} = 1;
    $self->writeservers();
    return $self;

}


# only for debugging and errorchecking
#
sub logobj {
    my ($self, $loglevel, $prefix, @objs) = @_;

    foreach my $obj (@objs) {
        my $line = debugobj($obj);
        $self->log($loglevel, "$prefix:\n $line");
    }
}


#
# Mail Object
#
# Main data type used by check and report is the Mail Object.
# an array of hash ref's, where array order matches mails in mbox (or stdin).
#
#             key = value  (not all defined) 
#
#              id = integer  NOTE: only key guaranteed to exist 
#       orig_mail = ref to string containing orig email (headers+body)
#         headers = headers of orig_email
#       bodyparts = bodyparts of orig_email, each maps to a sig in e2 (and e3, e4)
#     total_parts = number of bodyparts == num bodyparts_cleaned == num sigs in e1, etc.
#bodyparts_cleaned = sent through preprocessors (deHtml, deQP, etc..), debugging use only
#              e1 = hash of entire body (will always be a scalar)
#              e2 = array of hashes, 1st is for body part 1, 2nd for body 2, etc.
#              e3 = array of hashes, 1st is for headers, 2nd for body part 1, etc.
#              e4 = array of hashes, 1st is for headers, 2nd for body part 1, etc.
#            spam = 0, not spam, >1 spam
#       spamparts = spam rating for each body part. array of ints - 0|1  
#           known = 0|1  
#          skipme = 0|1  (not checked against server, usually whitelisted mail)
#            sent = array ref of hash ref's sent to server
#            resp = array ref of hash ref's, where hash is parsed sis of server response
#
sub prepare_objects {
    my ($self, $objs) = @_;
    my @objects;

    unless ($self->{s}->{engines} || 
            ($self->{s}->{engines} = $self->compute_supported_engines() ) ) {
        $self->log(1, "ALLBAD. supported engines not defined");
    }

    my $i = 1;
    foreach my $o (@$objs) {
        #
        #
        # fixme.  if email ends in \n\n, parse_mbox converts to \n\0
        # bug in mutt?
        #
        $self->log2file( 15, $o, "$i.before_prepare_obj" );
        my %obj = ( id => $i++ );
        if (ref($o) eq 'SCALAR') {  # checking/reporting mail
            $obj{orig_mail} = $o;
        } elsif (ref($o) eq 'HASH') { # checking cmd-line signatures
            $obj{"e$o->{eng}"} = $o->{sigs};
            $obj{ep4} = $o->{ep4} if $o->{ep4};
            $obj{total_parts} = scalar(@{$o->{sigs}});  # should always be 1
        }
        push @objects, \%obj;
    }
    return \@objects;
}


sub goodobjs {
    my ($self, $objects) = @_;

    # build list of good objs
    #
    my @objs;
    foreach my $obj (@$objects) {
        next if $obj->{skipme};
        push @objs, $obj;
    }
    $self->debug ("no valid dudes to check.  Done then.\n") unless @objs;
    return unless @objs;
    $self->log (6, "checking ". scalar(@objs) ." mail objs against the server ..");

    return \@objs;
}


# given mail objects, processes orig_mail and fills out 
#
# - headers 
# - bodyparts 
# - bodyparts_cleaned (debug use only)
# - e1
# - e2
# - e3
# - e4
#
# also returns array ref of sigs suitable for printing 
#
sub compute_sigs {
    my ($self, $objects) = @_;
    my @printable_sigs;

                            
    my $prep_mail_debug = 0; # debug print, 0=none, 1=split_mime stuff, 2=more verbose
    $prep_mail_debug++ if $self->{conf}->{debuglevel} > 15;
    $prep_mail_debug++ if $self->{conf}->{debuglevel} > 16;

    foreach my $obj (@$objects) {
        next unless $obj->{orig_mail};

        my ($subj) = ${$obj->{orig_mail}} =~ /\n(Subject: [^\n]+)\n/;
        $subj = substr $subj, 0, 50;
        $self->log(8,"mail ". $obj->{id} ." $subj");
        $self->log2file( 16, $obj->{orig_mail}, "$obj->{id}.orig_mail" );
        my ($headers, @bodyparts) = prep_mail (
                            $obj->{orig_mail}, 
                            $self->{conf}->{report_headers}, 
                            4 * 1024,
                            60 * 1024,
                            15 * 1024,
                            $self->{name_version},
                            $prep_mail_debug,  # $debug, 
                            );

        my $lines = " prep_mail done: mail $obj->{id} headers=". length($$headers);
        foreach (0..$#bodyparts) { $lines .= ", mime$_=". length(${$bodyparts[$_]}); }
        $self->log(8,$lines);

        unless (@bodyparts) {
            $self->log(2,"empty body in mail $obj->{id}, skipping");
            next;
        }
        $obj->{headers} = $headers;

        #$self->logobj(8,"obj id ". $obj->{id}, $obj);

        #
        # engine 1 is special case - must be same as razor 1
        #
        my ($hdrs, $body) = split /\n\r*\n/, ${$obj->{orig_mail}}, 2;
        $self->{preproc_vr1}->doit(\$body);
        my $sig = $self->compute_engine(1, \$body );
        $obj->{e1} = [$sig];
        push @printable_sigs, "$obj->{id} part N/A e1: $sig";
        

        # 
        # clean each bodypart, removing if new length is 0 
        #
        $obj->{bodyparts} = [];         
        $obj->{bodyparts_cleaned} = [];
        my $mime_num = -1;
        my $part = 0;
        foreach (@bodyparts) {
            my $olen = length($$_);
            my $clnpart = $$_;
            $mime_num++;
            $self->{preproc}->preproc( \$clnpart ); # in da future: $self->{s}->{conf}->{dre}
            if (length($clnpart) eq $olen) {
                $self->log(6,"preproc: mail $obj->{id} mime$mime_num unchanged, bytes=$olen, part $part");
            } elsif (length($clnpart) eq 0) {
                $self->log(6,"preproc: mail $obj->{id} mime$mime_num went from $olen bytes to 0, erasing");
                next;
            } else {
                $self->log(6,"preproc: mail $obj->{id} mime$mime_num went from $olen bytes to ". length($clnpart) .", part $part");
                #$self->logobj(5,"preproc: body before", $_);
                #$self->logobj(5,"preproc: body after", \$clnpart);
            }

            $self->log2file( 15, $_, "$obj->{id}.mime$mime_num.before.exactly_whats_in_msg_table" );
            $self->log2file( 15, \$clnpart, "$obj->{id}.mime$mime_num.after" );

            push @{$obj->{bodyparts}}, $_;
            push @{$obj->{bodyparts_cleaned}}, \$clnpart;
            $part++;
        }
        $obj->{total_parts} = scalar(@{$obj->{bodyparts}});

        # 
        # compute sig for bodyparts that are cleaned.
        #
        $obj->{ep4} = $self->{s}->{conf}->{ep4};
        $part = 0;
        foreach my $clnpart (@{$obj->{bodyparts_cleaned}}) {

            $self->log(6,"computing on mail $obj->{id} part $part, len ". length($$clnpart));
            foreach (sort keys %{$self->{s}->{engines}}) {

                next if $_ eq '1';  # engine 1 handled above

                my $sig = $self->compute_engine( $_, $clnpart,
                     $obj->{ep4} );  # if not engine 4, this is ignored
                $self->log(6,"doh.  no sig for mail $obj->{id} body $part eng $_") unless $sig;
                
                push @{$obj->{"e$_"}}, $sig;

                my $line = "$obj->{id} part $part e$_: $sig";
                $line .= ", ep4: $obj->{ep4}" if ($_ eq '4');
                push @printable_sigs, $line;
            }
        }
    }
    return \@printable_sigs;
}


#
# this function is the only one that has to be aware
# of razor protocol syntax.  (not including random logging)
# the hashes generated here are eventually sent to to_batched_query.
#
sub make_query {
    my ($self, $params) = @_;


    if ($params->{action} =~ /^check/)  {
                    
        my %query = (  a => 'c', 
                    e => $params->{eng},
                    s => $params->{sig},
                 );
        $query{ep4} = $params->{ep4} if $query{e} eq '4';
        return \%query;

    } elsif ($params->{action} =~ /^rcheck/) {

        my %query = (  a => 'r', 
                    e => $params->{eng},
                    s => $params->{sig},
                 );
        $query{ep4} = $params->{ep4} if $query{e} eq '4';
        return \%query;

    } elsif ($params->{action} =~ /(report|revoke)/) {

        #
        # fixme - should this be in String.pm ??
        #
        # prep_mail already truncated headers and body parts > 64K
        my @dudes;
        my $n = 0;
        while ($params->{bodyparts}->[$n]) {
            my $line = ${$params->{headers}};
            while ( $params->{bodyparts}->[$n] && 
                    (length(${$params->{bodyparts}->[$n]}) 
                    + length($line) < $self->{s}->{conf}->{bqs} * 1024)) {

                $self->log(11, "bqs=". ($self->{s}->{conf}->{bqs} * 1024) .  
                    " adding to line [len=". length($line) ."] part $n [len=".  
                    length(${$params->{bodyparts}->[$n]}) ."], total len=". 
                    (length(${$params->{bodyparts}->[$n]}) + length($line)) );

                $line .= "\r\n". ${$params->{bodyparts}->[$n++]};
            }
            push @dudes, $line;
        }
        my @queries;
        foreach (@dudes) {
            push @queries, {  a => $params->{action} eq 'report' ? 'r' : 'revoke', 
                        message => $_,
                 };
        }
        return @queries;

    }

}

#
# prepare queries in correct syntax for sending over network
#
sub obj2queries {
    my ($self, $objects, $action) = @_;
    my @queries = ();
    
    foreach my $obj (@$objects) {
        push @queries, @{$obj->{sent}}; 
    }

    $self->log(8,"preparing ". scalar(@queries) ." queries");
    my $qbatched = to_batched_query( \@queries, 
                                 $self->{s}->{conf}->{bql},
                                 $self->{s}->{conf}->{bqs},
                                 1);
    $self->log(8,"sending ". scalar(@$qbatched) ." batches");

    return $qbatched;
}


#
# Parse response syntax, add info to appropriate object
#
sub queries2obj { 
    my ($self, $objs, $responses, $action) = @_;

    my @resp;
    foreach (@$responses) {
        # from_batched_query wants "-" in beginning, but not ".\r\n" at end
        s/\.\r\n$//sg;   
        my $arrayref = from_batched_query($_);
        push @resp, @$arrayref;
    }

    #$self->logobj(8,"from_batched_query", \@resp);
    $self->log(12,"processing ". scalar(@resp) ." responses");

    my $n = 0;
    while (@resp) {
        my $obj = $objs->[$n++];
        return $self->error("more responses than mail objs!") unless $obj;

        foreach (@{$obj->{sent}}) {
            push @{$obj->{resp}}, shift @resp; 
            $self->log(12,"adding a resp to obj $obj->{id}");
        }
        #$self->logobj(15,"adding resp to obj", $obj);
    }
    return 1;
}


sub check_resp {
    my ($self, $me, $sent, $resp) = @_;

    if (exists $resp->{err}) { 
        $self->logobj(4,"$me: got err $resp->{err} for query", $sent);
        return 0;
    } 
    if ($resp->{p} eq '1') {
        if (exists $resp->{cf}) {
            if ($resp->{cf} < $self->{s}->{min_cf}) { 
                $self->log (6,"$me: Not spam: cf $resp->{cf} < min_cf $self->{s}->{min_cf}");
                return 0;
            } else {
                $self->log (6,"$me: Is spam: cf $resp->{cf} >= min_cf $self->{s}->{min_cf}");
                return 1;
            }
        }
        $self->log (6,"$me: sig found, no cf, ok.");
        return 1;
    }
    if ($resp->{p} eq '0') {
        $self->log (6,"$me: sig not found.");
        return 0;
    }
    # should never get here
    $self->logobj(2,"$me: got bad response from server - sent obj, resp obj", 
            [$sent, $resp] );
    return 0;
}


sub rcheck_resp {
    my ($self, $me, $sent, $resp) = @_;

    if (exists $resp->{err}) { 
        if ($resp->{err} eq '230') { 
            $self->log(8,"$me: err 230 - server wants mail");
            return 1;
        }
        $self->logobj(4,"$me: got err $resp->{err} for query", $sent);
        return 0;
    } 
    if ($resp->{res} eq '1') {
        $self->log (5,"$me: Server accepted report.");
        return 0;
    }
    if ($resp->{res} eq '0') {
        $self->log (1,"$me: Server did not accept report.  Shame on the server.");
        return 0;
    }
    # should never get here
    $self->logobj(2,"$me: got bad response from server - sent obj, resp obj", 
            [$sent, $resp] );
    return 0;
}


sub check { 
    my ($self, $objects) = @_;

    my $objs = $self->goodobjs($objects);  # skips those with skipme=1
    return 1 unless $objs;

    my $valid = 0;
    foreach my $obj (@$objs) {
        
        # Foreach obj, create array ref of hashes which are queries.
        # These will all be added to @queries which will be sent to_batched_query
        #
        $obj->{sent} = [];  

        #
        # Logic used in ordering of check queries 
        #
        # queries should go like this: (e=engine, p=bodypart)
        # e1, e2p0, e3p0, e4p0, e2p1, e3p1, e4p1
        # unless cmd-line sigs are passed.
        #
        foreach my $n (0..($obj->{total_parts} - 1)) {
            $self->log(8,"obj id ". $obj->{id} ." doing part $n");
            foreach (sort keys %{$self->supported_engines()}) {

                unless ($self->{s}->{engines}->{$_}) {
                    # warn if cmd-lig sig check is not supported
                    $self->log(5," engine $_ is not supported, sig check skipped") 
                        if ($obj->{"e$_"} && !$obj->{orig_mail});
                    next;
                }

                next unless $obj->{"e$_"}->[$n];  # engine 1 only has one hash

                $self->log(8," part $n, eng $_ sig: ". $obj->{"e$_"}->[$n] );

                my $query = $self->make_query( {
                             action => 'check',
                                sig => $obj->{"e$_"}->[$n],
                                ep4 => $obj->{ep4},
                                eng => $_ } );

                $valid++ if $query;
                push @{$obj->{sent}}, $query;
            }
        }
    }
    unless ($valid) {
        $self->log (5,"No queries, no spam");
        return 1;
    }


    # Build query text strings
    #
    my $queries = $self->obj2queries($objs, 'check') or return $self->errprefix("check 1");

    # send to server and store answers in mail obj
    #
    my $response = $self->_send($queries)           or return $self->errprefix("check 2");
    $self->queries2obj($objs, $response, 'check')   or return $self->errprefix("check 3");

    #
    # calculate spam
    #
    foreach my $obj (@$objects) {

    if ($obj->{skipme}) {
        $obj->{spam} = 0;
        next;
    }

    #
    # Logic for Spam
    #
    # Iterate thru responses.  If any is spam, mail is spam.
    #
    foreach (0..(scalar(@{$obj->{sent}}) - 1)) {
        my $sent = $obj->{sent}->[$_];
        my $resp = $obj->{resp}->[$_];
        unless ($resp) {
            $self->log(5,"doh. more sent queries than responses");
            next;
        }
        #my $me   = "mail ". $obj->{id} .", query $_, e=$sent->{e}";
        my $me   = "mail ". $obj->{id} .", query $_, e=$sent->{e} sig=$sent->{s}";
        $obj->{spam} += $self->check_resp($me, $sent, $resp);
    }


    $self->log (3,"mail id ". $obj->{id} ." is ". ($obj->{spam} ? '' : 'not ') ."known spam");

    }

    return 1;
} 



# returns hash ref if successfully registered
# returns 0 if not
sub register {
    my ($self, $p,) = @_;
    my @queries;

    my $registrar = $self->{name_version};
    my %qr = ( a => 'reg', registrar => $registrar );
    $qr{user} = $p->{user} if $p->{user};
    $qr{pass} = $p->{pass} if $p->{pass};
    $queries[0] = makesis(%qr);

    my $response = $self->_send(\@queries) or return $self->errprefix("register");

    my %resp = parsesis($$response[0]);
 
    return $self->error("Error $resp{err}: User exists. Try another name. aborting.\n")
        if ($resp{err} && $resp{err} eq '210'); 
    return $self->error("Error $resp{err} while performing register, aborting.\n")
        if ($resp{err});

    return $self->error("No success (res=$resp{res}) while performing register, aborting.\n")
        if ($resp{res} ne '1');

    $self->log(6,"Successfully registered with $self->{s}->{ip} identity: $resp{user}");

    # otherwise return hash containing 'user' and 'pass'
    delete $resp{res};
    return \%resp;
}


sub authenticate {
    my ($self, $options) = @_;
    my @queries;

    my %qr = ( a => 'ai', user => $options->{user} );
    $queries[0] = makesis(%qr);

    my $response = $self->_send(\@queries) or return $self->errprefix("authenticate 1");

    my %resp = parsesis($$response[0]);
    return $self->error("Error $resp{err} while authenticating, aborting.\n") if ($resp{err});
    my ($iv1, $iv2) = xor_key($options->{pass});
    my ($my_digest) = hmac_sha1($resp{achal}, $iv1, $iv2);

    %qr = ( a => 'auth', aresp => $my_digest );
    $queries[0] = makesis(%qr);

    $response = $self->_send(\@queries) or return $self->errprefix("authenticate 2");

    %resp = parsesis($$response[0]);
    return $self->error("Error $resp{err} while authenticating, aborting.\n") if ($resp{err});
    return $self->error("Authentication failed for user=$options->{user}")
        if ($resp{res} ne '1');

    $self->debug ("Authenticated user=$options->{user}");
    $self->{authenticated} = 1;
    return 1;
}


#
# handles report and revoke
#
sub report {
    my ($self, $objects) = @_;
    return $self->error("report: Not Authenticated") unless $self->{authenticated};

    return $self->error("report/revoke for engine 1 not supported")
            if ($self->{s}->{conf}->{dre} == 1);

    my @robjs;
    my $valid = 0;
    if ($self->{breed} eq 'report') {

        #
        # Before reporting entire email, check to see if server already has it
        #
    
        # skips those with skipme=1 ;
        my $objs = $self->goodobjs($objects);
        return 1 unless $objs;
    
        unless ($self->{s}->{conf}->{dre}) {
            $self->logobj(8,"server has no default dre, using 4", $self->{s}->{conf});
            $self->{s}->{conf}->{dre} = 4;
        }
    
        foreach my $obj (@$objs) {
            $obj->{sent} = [];

            # handle special case for engine  1
            # note: razor 1 does not store emails in its db, just sigs.
            # so we should never get a res=230 for e=1 a=r sig=xxx
            #
            push @{$obj->{sent}}, $self->make_query( {
                         action => 'rcheck',
                            sig => $obj->{e1}->[0],
                            eng => 1, } );
            foreach my $cur_bodypart (0..($obj->{total_parts} -1)) {
                push @{$obj->{sent}}, $self->make_query( {
                         action => 'rcheck',
                            sig => $obj->{"e$self->{s}->{conf}->{dre}"}->[$cur_bodypart],
                            ep4 => $obj->{ep4},
                            eng => $self->{s}->{conf}->{dre}, } );
            }
            $valid++ if scalar(@{$obj->{sent}});
            $self->log(12,"mail $obj->{id} ($obj->{total_parts} parts) sending ". 
                    scalar(@{$obj->{sent}}) ." queries");
        }
        unless ($valid) {
            $self->log (5,"No report check queries, no spam");
            return 1;
        }

        # Build query text strings - signatures computed already (see reportit)
        my $queries = $self->obj2queries($objs,'rcheck') or return $self->errprefix("report1");
    
        # send to server and store answers in mail obj
        my $response = $self->_send($queries)            or return $self->errprefix("report2");
        $self->queries2obj($objs, $response, 'rcheck')   or return $self->errprefix("report3");
                    
    
        $valid = 0;
        #
        # If server wants email or certain body parts, 
        # create new $obj->{sent} and add obj to @robjs
        #
        foreach my $obj (@$objs) {

            $self->log(12,"mail $obj->{id} read ". scalar(@{$obj->{resp}}) ." queries");
            # handle engine 1 special case
            if ( $self->rcheck_resp(
                    "mail ". $obj->{id} .", orig_email, special case eng 1",
                    $obj->{sent}->[0], $obj->{resp}->[0]) ) {
                $self->log(5,"doh.  Server should not send res=230 for eng=1 report");
            }

            my @bodyparts;
            my $cur_bodypart = -1;
            while ($obj->{bodyparts}->[++$cur_bodypart]) {
    
                next unless $self->rcheck_resp(
                    "mail ". $obj->{id} .", part $cur_bodypart, eng $self->{s}->{conf}->{dre}",
                    $obj->{sent}->[$cur_bodypart+1], $obj->{resp}->[$cur_bodypart+1]);

                push @bodyparts, $obj->{bodyparts}->[$cur_bodypart];
            }
            if (@bodyparts) {
                $obj->{resp} = [];  # clear responses from rcheck
                $obj->{sent} = [];  # clear responses from rcheck
                push @{$obj->{sent}}, $self->make_query( {
                                   action => 'report',
                                  headers => $obj->{headers},
                                bodyparts => \@bodyparts } );
                $valid++ if scalar(@{$obj->{sent}});
                push @robjs, $obj;
            }
        }

    } else {  # revoke

        foreach my $obj (@$objects) {
            $obj->{sent} = [];

            if (0) {  # don't revoke eng 1
            # handle special case for engine  1
            # note: razor 1 does not store emails in its db, just sigs.
            push @{$obj->{sent}}, $self->make_query( {
                         action => 'revoke',
                            sig => $obj->{e1}->[0],
                            eng => 1, } );
            } 
            # engines > 1 we send all the body parts
            push @{$obj->{sent}}, $self->make_query( {
                               action => 'revoke',
                              headers => $obj->{headers},
                            bodyparts => $obj->{bodyparts} } );
            push @robjs, $obj;
            $valid++ if scalar(@{$obj->{sent}});
        }

    }
    unless ($valid && scalar(@robjs)) {
        $self->log (3,"Finished $self->{breed}.");
        return 1;
    }

    #$self->logobj(8,"report objs", \@robjs);

    #
    # send server mails/body parts either
    # revoked, or requested if reporting
    #
    my $queries = $self->obj2queries( \@robjs,$self->{breed}) or return $self->errprefix("report4");
    my $response = $self->_send( $queries )                   or return $self->errprefix("report5");
    $self->queries2obj( \@robjs, $response,$self->{breed}) or return $self->errprefix("report6");

    # we just do this to log server's response
    #
    foreach my $obj (@robjs) {
        my $cur = -1;
        while ($obj->{sent}->[++$cur]) {
            $self->rcheck_resp(
                "$self->{breed}: mail $obj->{id}, $cur",
                $obj->{sent}->[$cur], $obj->{resp}->[$cur]);
        }
    }
    $self->log (3,"Sent $self->{breed}.");
    return 1;
}



sub _send { 
    my ($self, $msg, $closesock, $skipread) = @_;
    
    # connect to server, check version etc.
    #
    unless ($self->{select}) { 
        $self->connect() or return $self->errprefix("_send");
    }

    my @response;
    my $select = $self->{select};
    my $sock  = ($select->handles)[0];
    foreach my $i (0 .. ((scalar @$msg) -1) ) {
        my @handles = $select->can_write (15);
        if ($handles[0]) {
            $self->log (4,"$self->{s}->{ip} << ". length($$msg[$i]) );
            if ($$msg[$i] =~ /message/) { 
                my $line = debugobj($$msg[$i]);
                $self->log (6, $line );
                my $s = $$msg[$i];
                $self->log2file(16, \$s, "report_sent_to_server.$i" );
            } else {
                $self->log (6, $$msg[$i] );
            }
            local $\;
            undef $\; 
            print $sock $$msg[$i];
        } else {
            return $self->error("Timed out (15 sec) while writing to $self->{s}->{ip}");
        }
        next if $skipread;
        @handles = $select->can_read (15);
        if ($sock=$handles[0]) {
            local $/;
            undef $/; 
            $response[$i] = $self->_read($sock) or return $self->error("Error reading socket");
            $self->log (4,"$self->{s}->{ip} >> ". length($response[$i]) );
            $self->log (6, $response[$i]); 
        } else {
            return $self->error("Timed out (15 sec) while reading from $self->{s}->{ip}");
        }
    }
    if ($closesock) {
        $select->remove($sock);
        close $sock;
    }
    return \@response;
}


sub _read {
    my ($self, $socket) = @_;
    my ($query, $read);

        # fixme - need to trim this down (copied from server)
        #

        unless ($read = sysread($socket, $query, 1024)) {

            # There was an error on sysread(), could be a real error or a
            # blocking error.

            if ($! == EWOULDBLOCK) {
                # write would block, so we try again later
                $self->debug ("_read: EWOULDBLOCK");
                return;
            } elsif ($! == EINTR or $! == EIO) {
                # sysread() got interupted by a signal.
                # we will process this socket on next wheelwalk.
                $self->debug ("_read: EINTR");
                return;
            } elsif ($! == EPIPE or $! == EISDIR or $! == EBADF or $! == EINVAL or $! == EFAULT) {
                $self->debug ("_read: EPIPE");
                return;
            } else {
                # This happens when client breaks the connection.
                # Find out why we don't get an EPIPE instead. FIX!
                $self->debug ("_read: connection_closed");
                return;
            }
        }

        if ($read > 0) {

            # Now we are absolutely sure there is data on the socket.

            return $query;

        } else {

            # Otherwise we got an EOF, expire the socket

            $self->debug ("_read: EOF, connection_closed");
            return;

        }
}



sub connect {

    my ($self, %params) = @_;
    my $sock;

    if ($self->{simulate}) { 
        return $self->error ("Razor Error 4: This is a simulation. Won't connect to $self->{s}->{ip}.");
    }
    my $server = $params{server} || $self->{s}->{ip};
    my $port   = $params{port}   || $self->{s}->{port};
    unless ($port =~ /^\d+$/) {
        $self->log (6, "No proper port specified $port, using 2703"); # bootstrap_discovery will come here
        $port  = 2703;
    }
    $self->log (5,"Connecting to $server ...");
    if (my $proxy = $self->{conf}->{proxy}) {
        # 
        # Proxy stuff never been tested 
        # 
        $proxy =~ s!^http://!!;
        $proxy =~ s!:(\d+)/?$!!;
        my $pport = $1 || 80;
        $self->debug ("HTTP tunneling through $proxy:$pport.");
        $sock = IO::Socket::INET->new( 
                            PeerAddr => $proxy,
                            PeerPort => $pport,
                            Proto    => 'tcp',
                            Timeout  => 20,
                     );
        unless ( $sock ) {
            $self->debug ("Unable to connect to proxy $proxy:$pport; Reason: $!.");
        } else {    
            $sock->printf( "CONNECT %s:%d HTTP/1.0\r\n\r\n", $server, $port );
            if( $sock->getline =~ m!^HTTP/1\.0 200 ! ){
                # Skip through remaining part of MIME header.
                while( $sock->getline !~ m!^\r! ){ ; }
            } else {
                $self->log (4, "HTTP tunneling is disabled at $proxy.");
                $sock = undef;
            }
        }
    }

    # if proxy, we already might have a $sock.
    # if proxy failed to connect, try without proxy.
    #
    unless ($sock) {
        $sock = IO::Socket::INET->new( 
                            PeerAddr => $server,
                            PeerPort => $port,
                            Proto    => 'tcp',
                            Timeout  => 20,
                     );
        unless ( $sock ) {
            $self->log (3,"Unable to connect to $server:$port; Reason: $!.");
            return if $params{discovery_server};
            $self->nextserver or return $self->errprefix("connect1");
            return $self->connect;
        }    
    } 

    my $select = new IO::Select ($sock);
    my @handles = $select->can_read (15);
    if ($handles[0]) {
        $self->log (8,"Connection established");
        my $greeting = <$sock>;
        # $sock->autoflush; # this is on by default as of IO::Socket 1.18
        $self->{sock} = $sock;
        $self->{select} = $select;
        $self->log(4,"$server >> ". length($greeting) ." server greeting: $greeting");

        return 1 if $params{discovery_server};
        unless ($self->parse_greeting($greeting) ) {
            $self->nextserver or return $self->errprefix("connect2");
            return $self->connect;
        }
        return 1;
    } else {
        $self->log (3, "Timed out (15 sec) while reading from $self->{s}->{ip}.");
        $select->remove($sock);
        $sock->close();
        return $self->errprefix("connect3") if $params{skip_greeting};
        $self->nextserver or return $self->errprefix("connect4");
        return $self->connect;
    }
}

sub disconnect {
    my $self = shift;

    return unless $self->{sock};

    $self->log (5,"disconnecting from server ". $self->{s}->{ip});

    my $response = $self->_send(["a=q\r\n"], 0, 1) or return $self->errprefix("disconnect");

    delete $self->{sock};  # _send closes socket

    return 1;
}


sub parse_greeting {
    my ($self, $greeting) = @_;

    my %server_greeting = parsesis($greeting); 
    $self->{s}->{greeting} = \%server_greeting;

    unless ($self->{s}->{greeting} && $self->{s}->{greeting}->{sn}) { 
        $self->log(1,"Couldn't parse server greeting\n"); 
        return;
    }

    # server greeting must contain: sn, srl
    # server greeting may  contain: ep4, redirect, a,

    #
    # fixme - add support for redirect, etc.
    #

    #
    # current server config info is stored in $self->{s}->{conf}
    # see nextserver for more info
    #
    # If server greeting says there are new values
    # (which we know if greeting's srl > conf's srl)
    # we ask server for new values, update conf, then
    # put that server on modified list so it gets recorded to disk
    #
    # fixme - in the future, we could have a key with no value
    #         in .conf file - forcing client to ask server 'a=g&pm=key' 
    #

    unless ($self->{s}->{greeting}->{srl} > $self->{s}->{conf}->{srl}) { 
        $self->compute_server_conf;
        return 1 ;
    }

    # srl > our cached srl, request update  (a=g&pm=state)
    # and rediscover
    #

    my @queries = ("a=g&pm=state\r\n");

    my $response = $self->_send(\@queries) or return $self->errprefix("parse_greeting");

    # should be just one response
    # from_batched_query wants "-" in beginning, but not ".\r\n" at end
    $response->[0] =~ s/\.\r\n$//sg;   
    my $h = from_batched_query($response->[0], {});

    foreach my $href (@$h) {
        foreach (sort keys %$href) {
            $self->{s}->{conf}->{$_} = $href->{$_};
            #$self->log(8,"updated: $_=$href->{$_}");
        } 
    } 
    $self->log(1,"Bad info while trying to get server state (a=g&pm=state)") 
        unless scalar(@$h);

    $self->{s}->{conf}->{srl} = $self->{s}->{greeting}->{srl};
    push @{$self->{s}->{modified}}, $self->{s}->{ip};
    $self->{s}->{allconfs}->{$self->{s}->{ip}} = $self->{s}->{conf}; # in case new server

    # now we're up to date
    $self->log(5,"Updated to new server state srl ". $self->{s}->{conf}->{srl}
                   ." for server ". $self->{s}->{ip});

    $self->compute_server_conf();
    $self->writeservers;  # writes to disk servers listed in $self->{s}->{modified}

    $self->log(5,"srl was updated, forcing discovery ...");
    $self->{done_discovery} = 0;
    $self->{force_discovery} = 1;
    $self->discover();
     
    return 1;
}


# Returns engines supported 
#
# can be called with no paramaters or 
# with hash of server supported engines
sub compute_supported_engines {
    my ($self, $orig) = @_;

    my %all;
    my $se = $self->supported_engines(); # local supported engines
    foreach (@{$self->{conf}->{use_engines}}) {
        if ($orig) {
            $all{$_} = 1 if (exists $se->{$_}) && (exists $orig->{$_});
        } else {
            $all{$_} = 1 if exists $se->{$_};
        }
    }
    if ($orig) {
        $self->log(8, "Computed supported_engines: ". join(' ', sort(keys %all)) );
    } else {
        $self->log(8, "Client supported_engines: ". join(' ', sort(keys %all)) );
    }
    return \%all;
}


# called when we need to parse server conf
# - after initial parse_greeting
# - if state (srl) changes 
# - when we switch to cached server conf info in nextserver
#
sub compute_server_conf {
    my ($self, $cached) = @_;

    #
    # compute a confindence (cf) from razor-agent.conf's 'min_cf'
    # and server's average confidence (ac)
    #
    # min_cf can be 'n', 'ac', 'ac + n', or 'ac - n'
    # where 'n' can be 1..100
    # 
    my $cf     = $self->{s}->{conf}->{ac}; # default is server's ac
    my $min_cf = $self->{conf}->{min_cf};
    $min_cf =~ s/\s//g;

    if ($min_cf =~ /^ac\+(\d+)$/) {
        $cf = $self->{s}->{conf}->{ac} + $1;

    } elsif ($min_cf =~ /^ac-(\d+)$/) {
        $cf = $self->{s}->{conf}->{ac} - $1;

    } elsif ($min_cf =~ /^ac$/) {
        $cf = $self->{s}->{conf}->{ac};

    } elsif ($min_cf =~ /^(\d+)$/) {
        $cf = $min_cf;
    } else {
        $self->log(5,"Invalid min_cf $self->{conf}->{min_cf}");
    }
    $cf = 100 if $cf > 100;
    $cf = 0   if $cf < 0;
    $self->{s}->{min_cf} = $cf;

    #
    # ep4 - special for vr4
    #
    $self->{s}->{conf}->{ep4} = $self->{s}->{greeting}->{ep4} 
                             if $self->{s}->{greeting}->{ep4};

    my $info = $cached ? $self->{s}->{conf} : $self->{s}->{greeting};
    my $name = "Unknown-Type: ". $info->{sn};  
    $name = "Nomination" if $info->{sn} eq 'N';
    $name = "Catalogue"  if $info->{sn} eq 'C';

    $self->log (6, $self->{s}->{ip} ." is a $name Server srl ". 
        $self->{s}->{conf}->{srl} ."; computed min_cf=$cf, Server se: $self->{s}->{conf}->{se}");

    #
    # Supported Engines - greeting contains hex of bits
    # we turn into a hash so we can just quickly do
    # do_eng3_stuff if $self->{s}->{engines}->{3};
    #

    # if we're just computing hashes locally, ignore what engines server currently supports
    # fixme - this prolly should be done somewhere else
    if ($self->{opt}->{printhash}) {
        $self->log (6, "Ignore what engines server supports for -H");
        $self->{s}->{engines} = $self->compute_supported_engines();
    } else { 
        my $se = hexbits2hash($self->{s}->{conf}->{se} );
        $self->{s}->{engines} = $self->compute_supported_engines($se);
    } 
}


sub log2file_delme { 
    my ($self, $msgref, $mailid) = @_;

    #return; # comment this out for debug mode

    my $len = length($$msgref);
    my $fn = "/tmp/.razor.debug.msg.$$.$mailid";
    if (open OUT, ">$fn") {
        print OUT $$msgref;
        close OUT;
        $self->log(5,"log2file: wrote message len=$len to file: $fn");
    } else { 
        $self->log(5,"log2file: could open $fn: $!");
    }
}

sub debug { 
    my ($self, $message) = @_;
    $self->log(5,$message);
}


sub DESTROY { 
    my $self = shift;
    #$self->debug ("Agent terminated");
}


sub zonename { 
    my ($zone, $type) = @_;
    my ($sub, $dom) = split /\./, $zone, 2; 
    return "$sub-$type.$dom";
}


1;


